program oper2

  use sep
  use mem_mod
  use fbi_mod

  implicit none

  logical                             :: verb,adj, sou_ghost, rec_ghost
  logical                             :: zero_bound,keep_source_axis
  logical                             :: no_source_path,no_receiver_path
  integer                             :: nt, n1, n2,n4, wfld_j, i, i_4
  integer                             :: n_recurs,recurs_type
  integer                             :: n_pad_bpass
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, nlag, olag, lag_type, geom_type
  real                                :: t0=0,dt, d1, o1, d2, o2, o4, d4
  real                                :: tdelay_wavelet
  real                                :: epsi_gradio
  real                                :: f0_low,f1_low
  real                                :: f0_high,f1_high

  integer                             :: oper

  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:), allocatable   :: vel, deltaB
  real, dimension(:,:,:), allocatable :: refl,refl_mult
  real, dimension(:,:), allocatable ::  source 
  real, dimension(:,:,:), allocatable :: data, deltaD, deltaI
  real, dimension(:,:,:,:), allocatable :: refl_4d,deltaI_4d
  real, dimension(:,:,:), allocatable   :: wavefield,temp1

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_param("oper", oper)
  if(oper < 0 .or. oper > 7) call erexit("oper value is not correct")

  call from_param("geom_type", geom_type, 1) ! geom_type=1 is fixed receivers (land), geom_type=2 is fixed offset (marine)
  if(geom_type < 1 .or. geom_type > 2) call erexit("geom_type value is not correct")


  if ((oper .eq. 5).or.(oper .eq. 6)) then
!    call from_aux("data","n1", n1)
!    call from_aux("data","o1", o1)
!    call from_aux("data","d1", d1)
!    call from_aux("data","n2", n2)
!    call from_aux("data","o2", o2)
!    call from_aux("data","d2", d2)
    call from_history("n1", n1)
    call from_history("o1", o1)
    call from_history("d1", d1)
    call from_history("n2", n2)
    call from_history("o2", o2)
    call from_history("d2", d2)
    call from_aux("data","n3", nt)
    call from_aux("data","o3", t0)
    call from_aux("data","d3", dt)
  else
    call from_history("n1", n1)
    call from_history("o1", o1)
    call from_history("d1", d1)
    call from_history("n2", n2)
    call from_history("o2", o2)
    call from_history("d2", d2)
    call from_param("nt", nt)
    call from_aux("wavelet", "d1", dt)
  end if
  call from_aux("wavelet", "tdelay", tdelay_wavelet)

  call from_param("bc1", bc1,1)
  call from_param("bc2", bc2,1)
  call from_param("bc3", bc3,1)
  call from_param("bc4", bc4,1)
  call from_param("wfld_j", wfld_j,1)
  if(bc1 < 0 .or. bc2 < 0 .or. bc3 < 0 .or. bc4 < 0) call erexit("bc values are not correct")
  if(wfld_j < 1) call erexit("wfld_j value is not correct")

  call from_param("sou_nx", sou_nx,1)
  call from_param("sou_ox", sou_ox,1)
  call from_param("sou_dx", sou_dx,1)
  call from_param("sou_z", sou_z,1)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou_ox value is not correct")
  if(sou_dx < 1) call erexit("sou_dx value is not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx,1)
  call from_param("rec_ox", rec_ox,1)
  call from_param("rec_dx", rec_dx,1)
  call from_param("rec_z", rec_z,1)
  if(rec_dx < 1) call erexit("rec_dx value is not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  if(geom_type .eq. 1) then
    if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
    if(rec_ox < 1+bc2 .or. (rec_nx-1)*rec_dx+rec_ox > n2-bc4) write(0,*) "WARNING: recording inside the absorbing boundaries"
  else if(geom_type .eq. 2) then
    if(rec_ox + sou_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox + (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("rec axis values are not correct")
    if(rec_ox + sou_ox < 1+bc2 .or. (rec_nx-1)*rec_dx+rec_ox + (sou_nx-1)*sou_dx+sou_ox > n2-bc4) write(0,*) "WARNING: recording inside the absorbing boundaries"
    if(rec_nx<(0-rec_ox)/rec_dx+1) write(0,*) "WARNING: zero offset is not included"
  end if

  call from_param("adj", adj)
  call from_param("verb", verb,.false.)
  call from_param("sou_ghost", sou_ghost,.false.)
  call from_param("rec_ghost", rec_ghost,.false.)

  call from_param("zero_bound", zero_bound,.false.)
  call from_param("no_source_path", no_source_path,.false.)
  call from_param("no_receiver_path", no_receiver_path,.false.)
  call from_param("epsi_gradio", epsi_gradio,0.)
  if((oper .eq. 1) .or. (oper .eq. 3)) then
    call from_param("keep_source_axis", keep_source_axis,.false.)
    if(keep_source_axis) then
      n4=sou_nx
      o4=sou_ox
      d4=sou_dx
    else
      n4=1
      o4=0
      d4=1
    end if
  end if
  if ((oper .eq. 3) .and. adj) then
    call from_aux("deltaI","n3", nlag)
    call from_aux("deltaI","o3", olag)
    call from_aux("deltaI","lag_type", lag_type)
    if(verb) write(0,*)'From DeltaI: nlag=',nlag,'olag=',olag,'lag_type=',lag_type
    if(nlag<1-olag) write(0,*) "WARNING: zero lag is not included"
  else
    call from_param("nlag", nlag, 1)
    call from_param("olag", olag, 0)
! lag_type=1 is subsurface offset, lag_type=2 is time shift
    call from_param("lag_type", lag_type, 1) 
    if(nlag<1-olag) write(0,*) "WARNING: zero lag is not included"
  end if
  if(lag_type < 1 .or. lag_type > 2)  then
    call erexit("lag_type value is not correct")
  end if

  if (oper .eq. 7) then
    call from_param("n_recurs", n_recurs,1)
    call from_param("recurs_type", recurs_type,0)
! for bandpassing
    call from_param("n_pad_bpass", n_pad_bpass,0)
    call from_param("f0_low_bpass", f0_low,0.)
    call from_param("f1_low_bpass", f1_low,0.)
    call from_param("f0_high_bpass", f0_high,1./(2.*dt))
    call from_param("f1_high_bpass", f1_high,1./(2.*dt))
  else
    n_recurs=0
  end if

  call fbi_init1(n1, o1, d1, n2, o2, d2, n4, bc1, bc2, bc3, bc4)
  call fbi_init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call fbi_init3(nt, dt, wfld_j, nlag, olag, lag_type, geom_type, sou_ghost, rec_ghost,verb)
  call fbi_init4(zero_bound,no_source_path,no_receiver_path,keep_source_axis, &
                 n_recurs,recurs_type,n_pad_bpass, &
                 f0_low,f1_low,f0_high,f1_high,epsi_gradio)

  if((oper .eq. 0) .or. (oper < 3 .and. .not. adj) .or. &
     (oper == 4 .and. .not. adj) .or. (oper .eq. 7)) then
    call to_history("n1", nt)
    call to_history("d1", dt)
    call to_history("o1", t0-tdelay_wavelet)
    call to_history("label1","Time (s)")

    call to_history("n2", rec_nx)
    call to_history("d2", rec_dx*d2)
    if(geom_type .eq. 1) then
      call to_history("o2", (rec_ox-1)*d2+o2)
      call to_history("label2","Receiver location (km)")
    else if(geom_type .eq. 2) then
      call to_history("o2", (rec_ox)*d2)
      call to_history("label2","Offset (km)")
    end if

    call to_history("n3", sou_nx)
    call to_history("d3", sou_dx*d2)
    call to_history("o3", (sou_ox-1)*d2+o2)
    call to_history("label3","Source location (km)")
  else
    call to_history("n1", n1)
    call to_history("o1", o1)
    call to_history("d1", d1)
    call to_history("n2", n2)
    call to_history("o2", o2)
    call to_history("d2", d2)
    if((oper .eq. 1) .or. (oper .eq. 3 .and. .not. adj))then
      call to_history("n3", nlag)
      call to_history("o3", olag)
      call to_history("d3", 1.)
      if(((oper .eq. 1) .or. (oper .eq. 3)).and. keep_source_axis)then
        call to_history("n4", sou_nx)
        call to_history("d4", sou_dx*d2)
        call to_history("o4", (sou_ox-1)*d2+o2)
        call to_history("label4","Source location (km)")
      end if
    end if
  end if


! If the tag "wfldxx" is specified as a parameter, the history file will be created and the binary file will be written

  if(exist_file("wflds0")) then
    call to_history("n1", n1, "wflds0")
    call to_history("o1", o1, "wflds0")
    call to_history("d1", d1, "wflds0")
    call to_history("n2", n2, "wflds0")
    call to_history("o2", o2, "wflds0")
    call to_history("d2", d2, "wflds0")
    call to_history("n3", nt, "wflds0")
    call to_history("o3", t0-tdelay_wavelet, "wflds0")
    call to_history("d3", dt, "wflds0")
    call to_history("label1","Depth (km)","wflds0")
    call to_history("label2","Horizontal loc. (km)","wflds0")
    call to_history("label3","Time (s)","wflds0")
  endif

  if(exist_file("wflds1")) then
    call to_history("n1", n1, "wflds1")
    call to_history("o1", o1, "wflds1")
    call to_history("d1", d1, "wflds1")
    call to_history("n2", n2, "wflds1")
    call to_history("o2", o2, "wflds1")
    call to_history("d2", d2, "wflds1")
    call to_history("n3", nt, "wflds1")
    call to_history("o3", t0-tdelay_wavelet, "wflds1")
    call to_history("d3", dt, "wflds1")
    call to_history("label1","Depth (km)","wflds1")
    call to_history("label2","Horizontal loc. (km)","wflds1")
    call to_history("label3","Time (s)","wflds1")
    if (oper .eq. 7) then
!      call to_history("n3", n_recurs)
!      call to_history("o3", 0)
!      call to_history("d3", 1)
!      call to_history("label3","Scattering order")

      call to_history("n4", n_recurs, "wflds0")
      call to_history("o4", 0, "wflds0")
      call to_history("d4", 1, "wflds0")
      call to_history("label4","Scattering order","wflds0")

      call to_history("n4", n_recurs, "wflds1")
      call to_history("o4", 0, "wflds1")
      call to_history("d4", 1, "wflds1")
      call to_history("label4","Scattering order","wflds1")
    end if
  endif

  if(exist_file("wfldr0")) then
    call to_history("n1", n1, "wfldr0")
    call to_history("o1", o1, "wfldr0")
    call to_history("d1", d1, "wfldr0")
    call to_history("n2", n2, "wfldr0")
    call to_history("o2", o2, "wfldr0")
    call to_history("d2", d2, "wfldr0")
    call to_history("n3", nt, "wfldr0")
    call to_history("o3", t0-tdelay_wavelet, "wfldr0")
    call to_history("d3", dt, "wfldr0")
    call to_history("label1","Depth (km)","wfldr0")
    call to_history("label2","Horizontal loc. (km)","wfldr0")
    call to_history("label3","Time (s)","wfldr0")
  endif

  if(exist_file("scatter_sou")) then
    write(0,*)"Output scatter_sou"
!    if(nlag .ne. 1) call seperr ("Output scatter_sou field only if nlag == 1")
    call to_history("n1", n1, "scatter_sou")
    call to_history("o1", o1, "scatter_sou")
    call to_history("d1", d1, "scatter_sou")
    call to_history("n2", n2, "scatter_sou")
    call to_history("o2", o2, "scatter_sou")
    call to_history("d2", d2, "scatter_sou")
    call to_history("n3", nt, "scatter_sou")
    if (adj) then
      call to_history("o3", t0-tdelay_wavelet+(nt-1)*dt, "scatter_sou")
      call to_history("d3", -dt, "scatter_sou")
    else
      call to_history("o3", t0-tdelay_wavelet, "scatter_sou")
      call to_history("d3", dt, "scatter_sou")
    end if
    call to_history("label1","Depth (km)","scatter_sou")
    call to_history("label2","Horizontal loc. (km)","scatter_sou")
    call to_history("label3","Time (s)","scatter_sou")
  endif

  if(exist_file("scatter_rec")) then
    write(0,*)"Output scatter_rec"
!    if(nlag .ne. 1) call seperr ("Output scatter_rec field only if nlag == 1")
    call to_history("n1", n1, "scatter_rec")
    call to_history("o1", o1, "scatter_rec")
    call to_history("d1", d1, "scatter_rec")
    call to_history("n2", n2, "scatter_rec")
    call to_history("o2", o2, "scatter_rec")
    call to_history("d2", d2, "scatter_rec")
    call to_history("n3", nt, "scatter_rec")
    if (adj) then
      call to_history("o3", t0-tdelay_wavelet+(nt-1)*dt, "scatter_rec")
      call to_history("d3", -dt, "scatter_rec")
    else
      call to_history("o3", t0-tdelay_wavelet, "scatter_rec")
      call to_history("d3", dt, "scatter_rec")
    end if
    call to_history("label1","Depth (km)","scatter_rec")
    call to_history("label2","Horizontal loc. (km)","scatter_rec")
    call to_history("label3","Time (s)","scatter_rec")
  endif

  if(exist_file("scatter_tot")) then
    write(0,*)"Output scatter_tot"
!    if(nlag .ne. 1) call seperr ("Output scatter_tot field only if nlag == 1")
    call to_history("n1", n1, "scatter_tot")
    call to_history("o1", o1, "scatter_tot")
    call to_history("d1", d1, "scatter_tot")
    call to_history("n2", n2, "scatter_tot")
    call to_history("o2", o2, "scatter_tot")
    call to_history("d2", d2, "scatter_tot")
!    call to_history("n3", nt, "scatter_tot")
    call to_history("n3", 2*nt, "scatter_tot")
    if (adj) then
      call to_history("o3", t0-tdelay_wavelet+(nt-1)*dt, "scatter_tot")
      call to_history("d3", -dt, "scatter_tot")
    else
      call to_history("o3", t0-tdelay_wavelet, "scatter_tot")
      call to_history("d3", dt, "scatter_tot")
    end if
    call to_history("label1","Depth (km)","scatter_tot")
    call to_history("label2","Horizontal loc. (km)","scatter_tot")
    call to_history("label3","Time (s)","scatter_tot")
  endif

  call mem_alloc(wavelet, nt)
  call mem_alloc(vel, n1, n2)

  call sep_read(vel)
  call sep_read(wavelet, "wavelet")

  if(verb) write(0,*)"minval(vel)=",minval(vel)
  if(verb) write(0,*)"maxval(vel)=",maxval(vel)

  call fbi_alloc(wavelet)
  call model_update2(vel)


  if(oper .eq. 0) then ! Nonlinear operator: source <=> full nonlinear data (equation 9)
!    call mem_alloc(source, nt, rec_nx, sou_nx)
    call mem_alloc(source, nt, sou_nx)
    call mem_alloc(data, nt, rec_nx, sou_nx)
    if(adj) then
      do i=1,sou_nx
        call sep_read(data(:,:,i), "data")
      end do
      call L_nlop2(adj, .false., source, data)
      do i=1,sou_nx
!        call sep_write(source(:,:,i))
        call sep_write(source(:,i))
      end do
    else
      do isou=1,sou_nx
        sou_x = (isou-1)*sou_dx + sou_ox
        if(geom_type .eq. 1) then
          irec = (sou_x-rec_ox)/rec_dx + 1
        else if(geom_type .eq. 2) then
          irec = (0-rec_ox)/rec_dx + 1
!          irec = (sou_x-rec_ox)/rec_dx + 1
        end if
!        source(:,irec,isou) = wavelet
        source(:,isou) = wavelet
      end do
      call L_nlop2(adj, .false., source, data)
      do i=1,sou_nx
        call sep_write(data(:,:,i))
      end do
    end if
  else if(oper .eq. 1) then ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
    call mem_alloc(refl_4d, n1, n2, nlag,n4)
    call mem_alloc(data, nt, rec_nx, sou_nx)
!write(0,*)'#1'
    if(adj) then
!write(0,*)'#2'
      do i=1,sou_nx
        call sep_read(data(:,:,i), "data")
      end do
!write(0,*)'#3'
      call L_lop2(adj, .false., refl_4d, data)
!write(0,*)'#4'
      do i_4=1,n4
        do i=1,nlag
          call sep_write(refl_4d(:,:,i,i_4))
        end do
      end do
!write(0,*)'#5'
    else
      do i_4=1,n4
        do i=1,nlag
          call sep_read(refl_4d(:,:,i,i_4), "refl")
        end do
      end do
      call L_lop2(adj, .false., refl_4d, data)
      do i=1,sou_nx
        call sep_write(data(:,:,i))
      end do
    end if
  else if(oper .eq. 2) then ! Tomographic operator: background change <=> Second order scattered data (i.e. data due to transmission changes) (equation 25)
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(deltaB, n1, n2)
    call mem_alloc(deltaD, nt, rec_nx, sou_nx)
    do i=1,nlag
      call sep_read(refl(:,:,i), "refl")
    end do
    if(adj) then
      do i=1,sou_nx
        call sep_read(deltaD(:,:,i), "deltaD")
      end do
      call T_lop2(adj, .false., deltaB, deltaD, refl)
      call sep_write(deltaB)
    else
      call sep_read(deltaB, "deltaB")
      call T_lop2(adj, .false., deltaB, deltaD, refl)
      do i=1,sou_nx
        call sep_write(deltaD(:,:,i))
      end do
    end if
  else if(oper .eq. 3) then ! WEMVA operator: background change <=> Perturbation change (equation 29)
    if(verb) write(0,*)'Apply WEMVA operator'
    call mem_alloc(data, nt, rec_nx, sou_nx)
    call mem_alloc(deltaB, n1, n2)
    call mem_alloc(deltaI_4d, n1, n2, nlag,n4)
    do i=1,sou_nx
      call sep_read(data(:,:,i), "data")
    end do
    if(adj) then
      do i_4=1,n4
        do i=1,nlag
          call sep_read(deltaI_4d(:,:,i,i_4), "deltaI")
        end do
      end do
      call W_lop2(adj, .false., deltaB, deltaI_4d, data)
      call sep_write(deltaB)
    else
      call sep_read(deltaB, "deltaB")
      call W_lop2(adj, .false., deltaB, deltaI_4d, data)
      do i_4=1,n4
        do i=1,nlag
          call sep_write(deltaI_4d(:,:,i,i_4))
        end do
      end do
    end if
  else if(oper .eq. 4) then ! Multiple scattering operator
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(refl_mult, n1, n2, nlag)
    call mem_alloc(data, nt, rec_nx, sou_nx)
    if(adj) then
      call erexit("Adjoint multiscattering is not implemented yet")
    else
      do i=1,nlag
        call sep_read(refl(:,:,i), "refl")
        call sep_read(refl_mult(:,:,i), "refl_mult")
      end do
      call M_lop2(adj, .false., refl, data, refl_mult)
      do i=1,sou_nx
        call sep_write(data(:,:,i))
      end do
    end if
  else if(oper .eq. 5) then ! Non linear gradiometry operator
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(wavefield, n1, n2, nt)
    if(adj) then
      call sep_read(wavefield, "data")
!write(0,*)"maxval(wavefield)=",maxval(wavefield)
      call Gradio_nlop2(adj, .false., refl, wavefield)
      do i=1,nlag
        call sep_write(refl(:,:,i))
      end do
    else
      call seperr("Forward non linear gradiometry is not implemented yet")
    end if
  else if(oper .eq. 6) then ! Linear gradiometry operator
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(wavefield, n1, n2, nt)
    call mem_alloc(temp1, n1, n2, nt)

    call sep_read(wavefield, "data")
    call wavefield_update2(wavefield)
    if(adj) then
      call wave_oper_op2(.false., .false.,wavefield, temp1) !
!write(0,*)"maxval(wavefield)=",maxval(wavefield)
      call Gradio_lop2(adj, .false., refl,temp1)
      do i=1,nlag
        call sep_write(refl(:,:,i))
      end do
    else
      call seperr("Forward linear gradiometry is not implemented yet")
    end if
   else if(oper .eq. 7) then ! Recursive multiple scattering
    call mem_alloc(refl, n1, n2, nlag)
    call mem_alloc(data, nt, rec_nx, sou_nx)
    if(adj) then
      call seperr("Adjoint multiple scattering by recursion is not implemented yet")
    else
      do i=1,nlag
        call sep_read(refl(:,:,i), "refl")
      end do
      call MRec_lop2(adj, .false., refl, data)
      do i=1,sou_nx
        call sep_write(data(:,:,i))
      end do
    end if
  end if

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

end program
