module fbi_mod

  use sep
  use mem_mod
  use tomo_mod
  use sscale_mod
  use prop_mod
  use scat_mod
  use escats_mod
  use escatr_mod
  use escat_refl_mod
  use secder_mod
  use laplacian_mod
  use wave_oper_mod
  use bandpass_mod
!  use laplacian_4th_mod

  implicit none
  logical, private                        :: zero_bound
  logical, private                        :: no_source_path,no_receiver_path
  logical, private                        :: keep_source_axis
  integer, private                        :: nt, n1, n2, n4, nt2, wfld_j
  integer, private                        :: n_recurs,recurs_type
  integer, private                        :: n_pad_bpass
  integer, private                        :: sou_nx, sou_ox, sou_dx, sou_z, isou, sou_x, off_x
  integer, private                        :: rec_nx, rec_ox, rec_dx, rec_z, irec, rec_x
  integer, private                        :: nlag, olag, lag_type, geom_type, it
  integer, private                        :: bc1, bc2, bc3, bc4
  real, private                           :: dt, d1, o1, d2, o2, dtw
  real,private                            :: f0_low,f1_low
  real,private                            :: f0_high,f1_high
  real, private                           :: epsi_gradio
  logical, private                        :: verb,sou_ghost, rec_ghost
  real, dimension(:), allocatable, private             :: wavelet
  real, dimension(:,:), allocatable, private           :: vel, vel_mult, veldt2, vwt, wt
  double precision, dimension(:,:), allocatable, private   :: temp_space_lapl,temp_space_der2t
  real, dimension(:,:,:), allocatable, private         :: wfld0,wfld1, wfld2
  real, dimension(:,:,:), allocatable, private         :: wavefield
  real, dimension(:,:,:), allocatable, private         :: temp1, temp2
!  double precision, dimension(:,:,:), allocatable, private           :: temp_dbl
!  real, dimension(:,:,:), allocatable, private           :: temp_dbl
  contains

  subroutine fbi_init1(n1_in, o1_in, d1_in, n2_in, o2_in, d2_in, n4_in, bc1_in, bc2_in, bc3_in, bc4_in) ! Copying variables 1
    integer                             :: n1_in, n2_in, n4_in,  bc1_in, bc2_in, bc3_in, bc4_in
    real                                :: d1_in, o1_in, d2_in, o2_in
    n1 = n1_in
    n2 = n2_in
    d1 = d1_in
    o1 = o1_in
    d2 = d2_in
    o2 = o2_in
    n4 = n4_in
    bc1 = bc1_in
    bc2 = bc2_in
    bc3 = bc3_in
    bc4 = bc4_in
  end subroutine

  subroutine fbi_init2(sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in, rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in) ! Copying variables 2
    integer                             :: sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in
    integer                             :: rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in
    sou_nx = sou_nx_in
    sou_ox = sou_ox_in
    sou_dx = sou_dx_in
    sou_z = sou_z_in
    rec_nx = rec_nx_in
    rec_ox = rec_ox_in
    rec_dx = rec_dx_in
    rec_z = rec_z_in
  end subroutine

  subroutine fbi_init3(nt_in, dt_in, wfld_j_in, nlag_in, olag_in, lag_type_in, geom_type_in, sou_ghost_in, rec_ghost_in,verb_in) ! Copying variables 3
    integer                             :: nt_in, wfld_j_in, nlag_in, olag_in, lag_type_in, geom_type_in
    real                                :: dt_in
    logical                             :: sou_ghost_in, rec_ghost_in,verb_in
    nt = nt_in
    wfld_j = wfld_j_in
    dt = dt_in
    nlag = nlag_in
    olag = olag_in
    lag_type = lag_type_in
    geom_type = geom_type_in
    sou_ghost = sou_ghost_in
    rec_ghost = rec_ghost_in
    verb = verb_in
  end subroutine

  subroutine fbi_init4(zero_bound_in,no_source_path_in,no_receiver_path_in, &
                      keep_source_axis_in, &
                      n_recurs_in,recurs_type_in,n_pad_bpass_in,&
                      f0_low_in,f1_low_in,f0_high_in,f1_high_in,epsi_gradio_in)
    logical                     :: zero_bound_in,keep_source_axis_in
    logical                     :: no_source_path_in,no_receiver_path_in
    integer                     :: n_recurs_in
    integer                     :: recurs_type_in
    integer                     :: n_pad_bpass_in
    real                        :: f0_low_in,f1_low_in,f0_high_in,f1_high_in
    real                        :: epsi_gradio_in

    zero_bound=zero_bound_in
    no_source_path=no_source_path_in
    no_receiver_path=no_receiver_path_in
    keep_source_axis=keep_source_axis_in
    n_recurs=n_recurs_in
    recurs_type=recurs_type_in
    n_pad_bpass=n_pad_bpass_in
    f0_low=f0_low_in
    f1_low=f1_low_in
    f0_high=f0_high_in
    f1_high=f1_high_in
    epsi_gradio=epsi_gradio_in
  end subroutine

  subroutine fbi_alloc(wavelet_in) ! Allocations
    real, dimension(nt)                  :: wavelet_in
    integer                              :: one_int

    nt2 = (nt-1)*wfld_j+1
    dtw = dt / wfld_j
    one_int=1

    call mem_alloc(wavelet, nt)
    call mem_alloc(vel, n1, n2)
    call mem_alloc(vel_mult, n1, n2)
    call mem_alloc(veldt2, n1, n2)
    call mem_alloc(vwt, n1, n2)
    call mem_alloc(wt, n1, n2)
    call mem_alloc(wavefield, n1, n2, nt)
    call mem_alloc(temp1, n1, n2, nt)
    call mem_alloc(temp2, n1, n2, nt)
!    call mem_alloc(temp_dbl, n1, n2, nt)

    wavelet = wavelet_in

    call sscale_init(n1, n2, nt, dtw, veldt2, wt)
    call prop_init(n1, n2, nt, wfld_j, bc1, bc2, bc3, bc4, d1, d2, veldt2, vwt, wt)
    call scat_init(n1, n2, nt)
    call escats_init(n1, n2, nt, nlag, olag, lag_type)
    call escatr_init(n1, n2, nt, nlag, olag, lag_type)
    call escat_refl_init(n1, n2, nt, nlag, olag, lag_type)
    call secder_init(n1, n2, nt, dt,zero_bound)
    call laplacian_init(n1, n2, nt, d1, d2,zero_bound)
    call wave_oper_init(n1, n2, nt, vel,zero_bound)
!    call laplacian_4th_init(n1, n2, nt, d1, d2)
    if(n_recurs >1) then
     write(0,*)'Before bandpass_init'
      call bandpass_init(nt,n_pad_bpass,dt,f0_low,f1_low,f0_high,f1_high)
     write(0,*)'After bandpass_init'
    end if

  end subroutine

  function wavefield_update(wavefield_in) result(stat)
    integer             :: stat
    real, dimension(:)  :: wavefield_in
    call wavefield_update2(wavefield_in)
    stat=0
  end function

  subroutine wavefield_update2(wavefield_in) 
    real, dimension(n1, n2, nt)              :: wavefield_in
    wavefield= wavefield_in
  end subroutine

  function model_update(vel_in) result(stat)
    integer             :: stat
    real, dimension(:)  :: vel_in
    call model_update2(vel_in)
    stat=0
  end function

  subroutine model_update2(vel_in) ! Copy velocity model and prepare boundary condition stuff
    real, dimension(n1, n2)              :: vel_in

    vel = 0.
    vwt = 0.
    wt = 0.
    vel = vel_in
    call AbcWt(vel, vwt, wt, dtw, d1, d2, bc1, bc2, bc3, bc4)
    veldt2 = vel*vel*dtw*dtw
write(0,*)"In model_update: minval(vel)=",minval(vel)
write(0,*)"In model_update: maxval(vel)=",maxval(vel)
  end subroutine

  function L_nlop(adj, add, model, data) result(stat) ! Nonlinear operator fuction
    integer             :: stat
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data
    call L_nlop2(adj, add, model, data)
    stat=0
  end function

  function L_lop(adj, add, model, data) result(stat) ! Born operator fuction
    integer             :: stat 
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data
    call L_lop2(adj, add, model, data)
    stat=0
  end function

  function T_lop(adj, add, model, data, refl) result(stat) ! Tomographic operator fuction
    optional            :: refl
    integer             :: stat 
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data, refl
    if(.not. present(refl)) call seperr("optional parameter refl is required")
    call T_lop2(adj, add, model, data, refl)
    stat=0
  end function

  function W_lop(adj, add, model, data, dobs) result(stat) ! WEMVA operator fuction
    optional            :: dobs
    integer             :: stat 
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data, dobs
    if(.not. present(dobs)) call seperr("optional parameter dobs is required")
    call W_lop2(adj, add, model, data, dobs)
    stat=0
  end function

  function M_lop(adj, add, model, data, refl_mult) result(stat) ! Multiscattering operator fuction
    optional            :: refl_mult
    integer             :: stat 
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data, refl_mult
    if(.not. present(refl_mult)) call seperr("optional parameter refl_mult is required")
    call M_lop2(adj, add, model, data, refl_mult)
    stat=0
  end function

  function MRec_lop(adj, add, model, data) result(stat) ! Multiscattering operator fuction
    integer             :: stat 
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data
    call MRec_lop2(adj, add, model, data)
    stat=0
  end function

  function Gradio_nlop(adj, add, model, data) result(stat) ! Nonlinear operator fuction
    integer             :: stat
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data
    call Gradio_nlop2(adj, add, model, data)
    stat=0
  end function

  function Gradio_lop(adj, add, model, data) result(stat) ! Linear operator fuction
    integer             :: stat
    logical, intent(in) :: adj, add 
    real, dimension(:)  :: model, data
    call Gradio_lop2(adj, add, model, data)
    stat=0
  end function

  subroutine L_nlop2(adj, add, model, data) ! Nonlinear operator: source <=> full nonlinear data (equation 9)
    logical, intent(in)                  :: adj, add
!    real, dimension(nt, rec_nx, sou_nx)  :: model
    real, dimension(nt, sou_nx)  :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    if(adj) then ! Adjoint: full nonlinear data => source
      if(.not. add) model = 0.
      do isou=1,sou_nx
        sou_x = (isou-1)*sou_dx + sou_ox
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(rec_ghost) then ! K operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = data(:,irec,isou)
            temp1(rec_z+1,rec_x,:) = -data(:,irec,isou)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = data(:,irec,isou)
          end do
        end if
        call prop_op2(.true., .false., temp2, temp1) ! LG*L* operator
        call sscale_op2(.true., .false., temp1, temp2) ! S* operator
        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wflds0")
          end do
        end if
        if(sou_ghost) then ! K* operator
!          do irec=1,rec_nx
!            rec_x = (irec-1)*rec_dx + rec_ox + off_x
!            model(:,irec,isou) = model(:,irec,isou) + temp1(sou_z,rec_x,:)
!            model(:,irec,isou) = model(:,irec,isou) - temp1(sou_z+1,rec_x,:)
            model(:,isou) = model(:,isou) + temp1(sou_z,sou_x,:)
            model(:,isou) = model(:,isou) - temp1(sou_z+1,sou_x,:)
!          end do
        else
!          do irec=1,rec_nx
!            rec_x = (irec-1)*rec_dx + rec_ox + off_x
!            model(:,irec,isou) = model(:,irec,isou) + temp1(sou_z,rec_x,:)
            model(:,isou) = model(:,isou) + temp1(sou_z,sou_x,:)
!          end do
        end if
      end do
    else ! Forward: source => full nonlinear data
      if(.not. add) data = 0.
      do isou=1,sou_nx
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
!          do irec=1,rec_nx
!            rec_x = (irec-1)*rec_dx + rec_ox + off_x
!            temp1(sou_z,rec_x,:) = model(:,irec,isou)
!            temp1(sou_z+1,rec_x,:) = -model(:,irec,isou)
            temp1(sou_z,sou_x,:) = model(:,isou)
            temp1(sou_z+1,sou_x,:) = -model(:,isou)
!          end do
        else
!          do irec=1,rec_nx
!            rec_x = (irec-1)*rec_dx + rec_ox + off_x
!            temp1(sou_z,rec_x,:) = model(:,irec,isou)
            temp1(sou_z,sou_x,:) = model(:,isou)
!          end do
        end if
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, temp1) ! LGL* operator
        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wflds0")
          end do
        end if
        if(rec_ghost) then ! K* operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
            data(:,irec,isou) = data(:,irec,isou) - temp1(rec_z+1,rec_x,:)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
          end do
        end if
      end do
    end if
  end subroutine

  subroutine L_lop2(adj, add, model, data) ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2, nlag, n4)          :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    if(adj) then ! Adjoint: First order scattered data => perturbation
      call mem_alloc(wfld1, n1, n2, nt)
      if(.not. add) model = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, wfld1) ! LGL* operator

        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(wfld1(:,:,it), "wflds0")
          end do
        end if

        ! operator starts here
        temp1 = 0.
        if(rec_ghost) then ! K operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = data(:,irec,isou)
            temp1(rec_z+1,rec_x,:) = -data(:,irec,isou)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = data(:,irec,isou)
          end do
        end if
        call prop_op2(.true., .false., temp2, temp1) ! LG*L* operator
        call sscale_op2(.true., .false., temp1, temp2) ! S* operator
        if(exist_file("wfldr0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wfldr0")
          end do
        end if
        call secder_op2(.true., .false., temp2, temp1) ! T* operator
        ! (U^_s0)* operator
        if(keep_source_axis) then
          call escats_op2(.true., .false., model(:,:,:,isou), temp2, wfld1)
        else
          call escats_op2(.true., .true., model(:,:,:,1), temp2, wfld1)
        end if
      end do
      call mem_dealloc(wfld1)
    else ! Forward: perturbation => First order scattered data
      if(.not. add) data = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, temp1) ! LGL* operator

        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wflds0")
          end do
        end if

        ! operator starts here
        ! U^_s0 operator
        if(keep_source_axis) then
          call escats_op2(.false., .false., model(:,:,:,isou), temp2, temp1) 
        else
          call escats_op2(.false., .false., model(:,:,:,1), temp2, temp1)
        end if
        call secder_op2(.false., .false., temp2, temp1) ! T operator
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, temp1) ! LGL* operator

        if(exist_file("wflds1") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wflds1")
          end do
        end if

        if(rec_ghost) then ! K* operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
            data(:,irec,isou) = data(:,irec,isou) - temp1(rec_z+1,rec_x,:)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
          end do
        end if
      end do
    end if
  end subroutine

  subroutine T_lop2(adj, add, model, data, refl) ! Tomographic operator: background change <=> Second order scattered data (i.e. data due to transmission changes), needs perturbation
    optional                             :: refl
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2)              :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    real, dimension(n1, n2, nlag)        :: refl
    if(.not. present(refl)) call seperr("optional parameter refl is required")
    call mem_alloc(wfld1, n1, n2, nt)
    if(adj) then ! Adjoint: Second order scattered data => background change
      call mem_alloc(wfld2, n1, n2, nt)
      if(.not. add) model = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, wfld1) ! LGL* operator

        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(wfld1(:,:,it), "wflds0")
          end do
        end if

        ! Compute the u_r0 wavefield using the second order scattered data (this is a shared part of both halves of the operator)
        temp2 = 0.
        if(rec_ghost) then ! K operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp2(rec_z,rec_x,:) = data(:,irec,isou)
            temp2(rec_z+1,rec_x,:) = -data(:,irec,isou)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp2(rec_z,rec_x,:) = data(:,irec,isou)
          end do
        end if
        call prop_op2(.true., .false., temp1, temp2) ! LG*L* operator
        call sscale_op2(.true., .false., temp2, temp1) ! S* operator
        call secder_op2(.true., .false., wfld2, temp2) ! T* operator

        if(exist_file("wfldr0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(wfld2(:,:,it), "wfldr0")
          end do
        end if

        call escat_refl_op2(.true., .false., temp2, wfld2, refl) ! R* operator
        call prop_op2(.true., .false., temp1, temp2) ! LG*L* operator
        call sscale_op2(.true., .false., temp2, temp1) ! S* operator
        call secder_op2(.true., .false., temp1, temp2) ! T* operator
        call scat_op2(.true., .true., model, temp1, wfld1) ! (U_s0)* operator

        ! Precompute the u_ds wavefield using u_s0 wavefield then do the second half of the operator
        call escat_refl_op2(.false., .false., wfld1, temp2, refl) ! R operator
        call secder_op2(.false., .false., temp2, temp1) ! T operator
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, temp1) ! LGL* operator
        call scat_op2(.true., .true., model, temp1, wfld2) ! (U_ds)* operator
      end do
      call mem_dealloc(wfld2)
    else ! Forward: background change => Second order scattered data
      if(.not. add) data = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, wfld1) ! LGL* operator

        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(wfld1(:,:,it), "wflds0")
          end do
        end if

        ! Precompute the u_ds wavefield using u_s0 wavefield then do the first half of the operator
        call escat_refl_op2(.false., .false., wfld1, temp1, refl) ! R operator
        call secder_op2(.false., .false., temp1, temp2) ! T operator
        call sscale_op2(.false., .false., temp2, temp1) ! S operator
        call prop_op2(.false., .false., temp1, temp2) ! LGL* operator
        call scat_op2(.false., .false., model, temp1, temp2) ! U_ds operator
        call secder_op2(.false., .false., temp1, temp2) ! T operator
        call sscale_op2(.false., .false., temp2, temp1) ! S operator
        call prop_op2(.false., .false., temp1, temp2) ! LGL* operator
        if(rec_ghost) then ! K* operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp2(rec_z,rec_x,:)
            data(:,irec,isou) = data(:,irec,isou) - temp2(rec_z+1,rec_x,:)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp2(rec_z,rec_x,:)
          end do
        end if

        ! second half of operator starts here
        call scat_op2(.false., .false., model, temp1, wfld1) ! U_s0 operator
        call secder_op2(.false., .false., temp1, temp2) ! T operator
        call sscale_op2(.false., .false., temp2, temp1) ! S operator
        call prop_op2(.false., .false., temp1, temp2) ! LGL* operator
        call escat_refl_op2(.false., .false., temp2, temp1, refl) ! R operator
        call secder_op2(.false., .false., temp1, temp2) ! T operator
        call sscale_op2(.false., .false., temp2, temp1) ! S operator
        call prop_op2(.false., .false., temp1, temp2) ! LGL* operator
        if(rec_ghost) then ! K* operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp2(rec_z,rec_x,:)
            data(:,irec,isou) = data(:,irec,isou) - temp2(rec_z+1,rec_x,:)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp2(rec_z,rec_x,:)
          end do
        end if
      end do
    end if
    call mem_dealloc(wfld1)
  end subroutine

  subroutine W_lop2(adj, add, model, data, dobs) ! WEMVA operator: background change <=> Perturbation change, needs first order scattered data
    optional                             :: dobs
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2)              :: model
    real, dimension(n1, n2, nlag,n4)     :: data
    real, dimension(nt, rec_nx, sou_nx)  :: dobs
    if(.not. present(dobs)) call seperr("optional parameter dobs is required")
    call mem_alloc(wfld1, n1, n2, nt)
    call mem_alloc(wfld2, n1, n2, nt)
    if(adj) then ! Adjoint: Perturbation change => background change
      if(.not. add) model = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, wfld1) ! LGL* operator

        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(wfld1(:,:,it), "wflds0")
          end do
        end if

        ! Precompute the u_r0 wavefield using the first order scattered data
        temp1 = 0.
        if(rec_ghost) then ! K operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = dobs(:,irec,isou)
            temp1(rec_z+1,rec_x,:) = -dobs(:,irec,isou)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = dobs(:,irec,isou)
          end do
        end if
        call prop_op2(.true., .false., temp2, temp1) ! LG*L* operator
        call sscale_op2(.true., .false., temp1, temp2) ! S* operator
        call secder_op2(.true., .false., wfld2, temp1) ! T* operator

        if(exist_file("wfldr0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(wfld2(:,:,it), "wfldr0")
          end do
        end if

        if(.not. no_source_path) then
        ! operator starts here
        ! U^_r0 operator
          if(keep_source_axis) then
            call escatr_op2(.false., .false., data(:,:,:,isou), temp2, wfld2) 
          else
            call escatr_op2(.false., .false., data(:,:,:,1), temp2, wfld2) 
          end if
          call prop_op2(.true., .false., temp1, temp2) ! LG*L* operator
          call sscale_op2(.true., .false., temp2, temp1) ! S* operator
          call secder_op2(.true., .false., temp1, temp2) ! T* operator
          call scat_op2(.true., .true., model, temp1, wfld1) ! (U_s0)* operator
        end if

        if(.not. no_receiver_path) then
        ! second half of operator starts here
        ! U^_s0 operator
          if(keep_source_axis) then
            call escats_op2(.false., .false., data(:,:,:,isou), temp2, wfld1) 
          else
            call escats_op2(.false., .false., data(:,:,:,1), temp2, wfld1) 
          end if
          call secder_op2(.false., .false., temp2, temp1) ! T operator
          call sscale_op2(.false., .false., temp1, temp2) ! S operator
          call prop_op2(.false., .false., temp2, temp1) ! LGL* operator
  !        if(exist_file("scatter_rec")) then
  !          call sep_write(temp2, "scatter_rec")
  !        end if
          call scat_op2(.true., .true., model, temp1, wfld2) ! (U_r0)* operator
        end if

      end do
    else ! Forward: background change => Perturbation change
      if(.not. add) data = 0.
      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, wfld1) ! LGL* operator

        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(wfld1(:,:,it), "wflds0")
          end do
        end if

        ! Precompute the u_r0 wavefield using the first order scattered data
        temp1 = 0.
        if(rec_ghost) then ! K operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = dobs(:,irec,isou)
            temp1(rec_z+1,rec_x,:) = -dobs(:,irec,isou)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            temp1(rec_z,rec_x,:) = dobs(:,irec,isou)
          end do
        end if
        call prop_op2(.true., .false., temp2, temp1) ! LG*L* operator
        call sscale_op2(.true., .false., temp1, temp2) ! S* operator
        if(exist_file("wfldr0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wfldr0")
          end do
        end if
        call secder_op2(.true., .false., wfld2, temp1) ! T* operator

        if(.not. no_receiver_path) then
        ! operator starts here
          call scat_op2(.false., .false., model, temp1, wfld2) ! U_r0 operator
          call prop_op2(.true., .false., temp2, temp1) ! LG*L* operator
          call sscale_op2(.true., .false., temp1, temp2) ! S* operator
          if(exist_file("scatter_rec")) then
            call sep_write(temp2, "scatter_rec")
          end if
          call secder_op2(.true., .false., temp2, temp1) ! T* operator
          ! (U^_s0)* operator
          if(keep_source_axis) then
            call escats_op2(.true., .false., data(:,:,:,isou), temp2, wfld1)
          else
            call escats_op2(.true., .true., data(:,:,:,1), temp2, wfld1)
          end if
        end if

        if(.not. no_source_path) then
        ! second half of operator starts here
          call scat_op2(.false., .false., model, temp1, wfld1) ! U_s0 operator
          call secder_op2(.false., .false., temp1, temp2) ! T operator
          call sscale_op2(.false., .false., temp2, temp1) ! S operator
          call prop_op2(.false., .false., temp1, temp2) ! LGL* operator
          if(exist_file("scatter_sou")) then
            call sep_write(temp2, "scatter_sou")
          end if
  !! probable bug
  !!        call escatr_op2(.true., .true., data, temp1, wfld2) ! (U^_r0)* operator
           ! (U^_r0)* operator
          if(keep_source_axis) then
            call escatr_op2(.true., .false., data(:,:,:,isou), temp2, wfld2)
          else
            call escatr_op2(.true., .true., data(:,:,:,1), temp2, wfld2)
          end if
        end if
      end do
    end if
    call mem_dealloc(wfld1)
    call mem_dealloc(wfld2)
  end subroutine

  subroutine M_lop2(adj, add, model, data, refl_mult) ! perturbation <=> Multiscattered data, needs multiscattering reflectivity
    optional                             :: refl_mult
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2)              :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    real, dimension(n1, n2, nlag)        :: refl_mult
    if(.not. present(refl_mult)) call seperr("optional parameter refl_mult is required")
    if(adj) then ! Adjoint: Multiscattered data => perturbation
      call seperr("Adjoint multiscattering is not implemented yet")
    else ! Forward: perturbation => Multiscattered data
      if(.not. add) data = 0.
write(0,*)'#1'
      ! Compute the multiscattering velocity (using only the zero lag of refl_mult, which is in units of slowness squared), conversion is to match the units
      vel_mult = 1./(vel*vel)
      vel_mult = vel_mult + refl_mult(:,:,1-olag)
      vel_mult = 1./sqrt(vel_mult)
write(0,*)'#2'

      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
write(0,*)'#3'
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, temp1) ! LGL* operator
write(0,*)'#4'

        if(exist_file("wflds0") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wflds0")
          end do
        end if

        ! Use vel for the background wavefield
        call model_update2(vel_mult)
write(0,*)'#6'

        ! operator starts here
        call escats_op2(.false., .false., model, temp2, temp1) ! U^_s0 operator
        call secder_op2(.false., .false., temp2, temp1) ! T operator
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, temp1) ! LGL* operator

write(0,*)'#7'
        if(exist_file("wflds1") .and. isou .eq. 1) then
          do it=1,nt
            call sep_write(temp1(:,:,it), "wflds1")
          end do
        end if
write(0,*)'#8'

        if(rec_ghost) then ! K* operator
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
            data(:,irec,isou) = data(:,irec,isou) - temp1(rec_z+1,rec_x,:)
          end do
        else
          do irec=1,rec_nx
            rec_x = (irec-1)*rec_dx + rec_ox + off_x
            data(:,irec,isou) = data(:,irec,isou) + temp1(rec_z,rec_x,:)
          end do
        end if

write(0,*)'#9'

        ! Reset to the original velocity model
        call model_update2(vel)

      end do
    end if
  end subroutine

  subroutine MRec_lop2(adj, add, model, data) ! perturbation <=> Multiscattered data recursively updating background
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2)              :: model
    real, dimension(nt, rec_nx, sou_nx)  :: data
    integer i_recurs

write(0,*)'MRec - #0'
write(0,*)'n1=',n1
write(0,*)'n2=',n2
write(0,*)'nt=',nt

    if(recurs_type .eq. 1) then
      call mem_alloc(wfld0, n1, n2, nt)
    end if
    call mem_alloc(wfld1, n1, n2, nt)
    call mem_alloc(wfld2, n1, n2, nt)

write(0,*)'adj=',adj

    if(adj) then ! Adjoint: Multiscattered data => perturbation
      call seperr("Adjoint multiscattering is not implemented yet")
    else ! Forward: perturbation => Multiscattered data
write(0,*)'add=',add
      if(.not. add) data = 0.
write(0,*)'MRec - #1'

      do isou=1,sou_nx
        ! Precompute the u_s0 wavefield using the wavelet
        sou_x = (isou-1)*sou_dx + sou_ox
        if(verb) write(0,*)'sou_x=',sou_x
        off_x = 0
        if(geom_type .eq. 2) off_x = sou_x
        temp1 = 0.
        if(sou_ghost) then ! K operator
          temp1(sou_z,sou_x,:) = wavelet
          temp1(sou_z+1,sou_x,:) = -wavelet
        else
          temp1(sou_z,sou_x,:) = wavelet
        end if
write(0,*)'MRec - #2'
        call sscale_op2(.false., .false., temp1, temp2) ! S operator
        call prop_op2(.false., .false., temp2, wfld1) ! LGL* operator
write(0,*)'MRec - #3'

        if(recurs_type .eq. 1) then
          wfld0=wfld1
        end if

        do i_recurs=1,n_recurs

!          if(exist_file("wflds0").and.(isou .eq. 1).and.(i_recurs .eq. 1)) then
!          if(exist_file("wflds0").and.(isou .eq. 1)) then
!            do it=1,nt
!              call sep_write(wfld1(:,:,it), "wflds0")
!            end do
!          end if
  
!temp1=0
!temp2=0

        ! operator starts here
          call escats_op2(.false., .false., model, temp2, wfld1) ! 
          call secder_op2(.false., .false., temp2, temp1) ! T operator
          call sscale_op2(.false., .false., temp1, temp2) ! S operator
          call prop_op2(.false., .false., temp2, temp1) ! LGL* operator

          write(0,*)'Before bandpass condition'
          if(.not.((f0_low==0.).and.(f1_low==0.).and. &
                   (f0_high==1./(2.*dt)).and.(f1_high==1./(2.*dt)))) then
            write(0,*)'Before bandpass'
            call bandpass_3d3(temp1)
            write(0,*)'After bandpass'
          else
            write(0,*)'DO NOT Bandpass wavefield'
          end if

!!

          if(recurs_type .eq. 0) then
            wfld1=temp1
            wfld2=wfld2+temp1
          else
            wfld1=wfld0+temp1
            wfld2=temp1
         end if

write(0,*)'MRec - #4 - i_recurs=',i_recurs
          if(exist_file("wflds1") .and. isou .eq. 1) then
            do it=1,nt
              call sep_write(wfld2(:,:,it), "wflds1")
            end do
          end if
write(0,*)'MRec - #5 - i_recurs=',i_recurs
write(0,*)'MRec - #6 sum'

!          if(recurs_type .eq. 0) then
!            wfld1=wfld1+temp1
!          end if

         end do

          if(rec_ghost) then ! K* operator
            do irec=1,rec_nx
              rec_x = (irec-1)*rec_dx + rec_ox + off_x
              data(:,irec,isou) = data(:,irec,isou) + wfld2(rec_z,rec_x,:)
              data(:,irec,isou) = data(:,irec,isou) - wfld2(rec_z+1,rec_x,:)
            end do
          else
            do irec=1,rec_nx
              rec_x = (irec-1)*rec_dx + rec_ox + off_x
              data(:,irec,isou) = data(:,irec,isou) + wfld2(rec_z,rec_x,:)
            end do
          end if

write(0,*)'MRec - #6'

      end do
    end if
    call mem_dealloc(wfld1)
    call mem_dealloc(wfld2)
  end subroutine




  subroutine Gradio_nlop2(adj, add, model, data) 
    logical, intent(in)                  :: adj, add
    integer                   :: stat
    real, dimension(n1, n2, nlag)          :: model
    real, dimension(n1, n2, nt)  :: data

    if(adj) then 
      call mem_alloc(temp_space_lapl, n1, n2)
      call mem_alloc(temp_space_der2t, n1, n2)
      if(.not. add) model = 0.
        call laplacian_op2(.true., .false., temp1, data) ! LGL* operator
        temp_space_lapl=sum(temp1**2,dim=3)
        call secder_op2(.true., .false., temp2, data) ! LGL* operator
!        stat= secder_op(.true., .false., temp2, data) ! LGL* operator
        temp_space_der2t=sum(temp1*temp2,dim=3)
        model(:,:,1)=sqrt(abs(temp_space_der2t/(temp_space_lapl+epsi_gradio)))
    else ! Forward: perturbation => Multiscattered data
      call seperr("Forward non linear gradiometry is not implemented yet")
    end if

    call mem_dealloc(temp_space_der2t)
    call mem_dealloc(temp_space_lapl)

  end subroutine

  subroutine Gradio_lop2(adj, add, model, data) 
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2, nlag)          :: model
    real, dimension(n1, n2, nt)  :: data

    if(adj) then 
      if(.not. add) model = 0.

!write(0,*)"In Gradio_lop2 adj: minval(wavefield)=",minval(wavefield)
!write(0,*)"In Gradio_lop2 adj: maxval(wavefield)=",maxval(wavefield)
!write(0,*)"In Gradio_lop2 adj: minval(data)=",minval(data)
!write(0,*)"In Gradio_lop2 adj: maxval(data)=",maxval(data)

!        model(:,:,1)=sum(temp1(:,:,5:nt-5),dim=3)

        call secder_op2(.true., .false., temp2, wavefield) ! 
!        call laplacian_op2(.false., .false., wavefield,temp2) ! 
        call escats_op2(.true.,add, model, data, temp2) ! 

!        model=sqrt(abs(model))

!write(0,*)"In Gradio_lop2: minval(model)=",minval(model)
!write(0,*)"In Gradio_lop2: maxval(model)=",maxval(model)

    else ! Forward: 
      if(.not. add) data = 0.
!write(0,*)"In Gradio_lop2 fwdj: minval(wavefield)=",minval(wavefield)
!write(0,*)"In Gradio_lop2 fwdj: maxval(wavefield)=",maxval(wavefield)
!write(0,*)"In Gradio_lop2 fwd: minval(model)=",minval(model)
!write(0,*)"In Gradio_lop2 fwd: maxval(model)=",maxval(model)
!        call laplacian_op2(.false., .false., wavefield,temp2) ! 
!write(0,*)"In Gradio_lop2 fwd: minval(temp2)=",minval(temp2)
!write(0,*)"In Gradio_lop2 fwd: maxval(temp2)=",maxval(temp2)
        call secder_op2(.false., .false., wavefield,temp2) ! 
        call escats_op2(.false., add, model, data, temp2) ! 
!write(0,*)"After -In Gradio_lop2 fwdj: minval(wavefield)=",minval(wavefield)
!write(0,*)"After -In Gradio_lop2 fwdj: maxval(wavefield)=",maxval(wavefield)
!write(0,*)"After -In Gradio_lop2 fwdj: minval(data)=",minval(data)
!write(0,*)"After -In Gradio_lop2 fwdj: maxval(data)=",maxval(data)
    end if
  end subroutine


end module 
