module laplacian_4th_mod

  use mem_mod

  implicit none

  logical, private                    :: flat_bound
  integer, private                            :: n1, n2, nt
  integer, private                            :: fat
  integer, private                            :: iz, ix, it
  real, private                               :: stencil0
  real, dimension(5), private                 :: stencilz, stencilx
  real, dimension(:,:), allocatable, private  :: p0, p1

  contains

  subroutine laplacian_4th_init(n1_in, n2_in, nt_in, d1_in, d2_in,flat_bound_in)
    optional           :: flat_bound_in
    logical             :: flat_bound_in
    integer                        :: n1_in, n2_in, nt_in
    real                           :: d1_in, d2_in

    if(present(flat_bound_in)) then
      flat_bound=flat_bound_in
!  write(0,*)'true flat_bound=',flat_bound
    else
      flat_bound=.false.
!  write(0,*)'false flat_bound=',flat_bound
    end if

    n1 = n1_in
    n2 = n2_in
    nt = nt_in

!write(0,*)'In lapl init n1=',n1
!write(0,*)'In lapl init n2=',n2
!write(0,*)'In lapl init nt=',nt

    fat =1 

    stencilz(1)=2/(2.*d1_in*d1_in)

    stencilx(1)=2/(2.*d2_in*d2_in)

    stencil0=-4/(d2_in*d2_in)

!write(0,*)'stencil0=',stencil0

    if(.not. allocated(p0)) then
      allocate(p0(1-fat:n1+fat,1-fat:n2+fat))
      call mem_add_4(size(p0))
    end if
    if(.not. allocated(p1)) then
      allocate(p1(1-fat:n1+fat,1-fat:n2+fat))
      call mem_add_4(size(p1))
    end if

  end subroutine

  function laplacian_4th_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call laplacian_4th_op2(adj, add, model, data)
    stat=0
  end function

  subroutine laplacian_4th_op2(adj, add, model, data)
    logical,intent(in)        :: adj, add 
    real, dimension(n1,n2,nt) :: model, data

    integer                  :: ifat

    p1 = 0.

    if(adj) then
      if(.not. add) model = 0.
!write(0,*)"in lapl adjoint start with minval(data)=",minval(data)
!write(0,*)"in lapl adjoint start with maxval(data)=",maxval(data)
      do it=1,nt
        p0(1:n1,1:n2) = data(:,:,it)
        if(flat_bound) then
!write(0,*)' adjoint In flat_bound'
          do ifat=1,fat
            p0(1:n1,1-ifat) = data(:,1,it)
            p0(1:n1,n2+ifat) = data(:,n2,it)
            p0(1-ifat,1:n2) = data(1,:,it)
            p0(n1+ifat,1:n2) = data(n1,:,it)
            p0(1-ifat,1-ifat)=data(1,1,it)
            p0(1-ifat,n2+ifat)=data(1,n2,it)
            p0(n1+ifat,1-ifat)=data(n1,1,it)
            p0(n1+ifat,n2+ifat)=data(n1,n2,it)
          end do
        end if
        do ix=1,n2
          do iz=1,n1
            p1(iz,ix) =   stencil0*p0(iz,ix)+&
                          stencilz(1)*(p0(iz-1,ix  )+p0(iz+1,ix  ))+&
                          stencilx(1)*(p0(iz  ,ix-1)+p0(iz  ,ix+1))
          end do
        end do
        model(:,:,it) = model(:,:,it) + p1(1:n1,1:n2)
      end do
!write(0,*)"in lapl adjoint end with minval(model)=",minval(model)
!write(0,*)"in lapl adjoint end with maxval(model)=",maxval(model)
!call snap ("Lapla_adj.H",n1,n2,model)
    else
!write(0,*)"in lapl forward start with minval(mode)=",minval(model)
!write(0,*)"in lapl forward start with maxval(mode)=",maxval(model)
      if(.not. add) data = 0.
      do it=1,nt
        p0(1:n1,1:n2) = model(:,:,it)
        if(flat_bound) then
!write(0,*)' forward In flat_bound'
          do ifat=1,fat
            p0(1:n1,1-ifat) = model(:,1,it)
            p0(1:n1,n2+ifat) = model(:,n2,it)
            p0(1-ifat,1:n2) = model(1,:,it)
            p0(n1+ifat,1:n2) = model(n1,:,it)
            p0(1-ifat,1-ifat)=model(1,1,it)
            p0(1-ifat,n2+ifat)=model(1,n2,it)
            p0(n1+ifat,1-ifat)=model(n1,1,it)
            p0(n1+ifat,n2+ifat)=model(n1,n2,it)
          end do
        end if
        do ix=1,n2
          do iz=1,n1
            p1(iz,ix) =   stencil0*p0(iz,ix)+&
                          stencilz(1)*(p0(iz-1,ix  )+p0(iz+1,ix  ))+&
                          stencilx(1)*(p0(iz  ,ix-1)+p0(iz  ,ix+1))
          end do
        end do
        data(:,:,it) = data(:,:,it) + p1(1:n1,1:n2)
      end do
!write(0,*)"in lapl forward end with minval(data)=",minval(data)
!write(0,*)"in lapl forward end with maxval(data)=",maxval(data)
!call snap ("Lapla_forw.H",n1,n2,data)
    end if
  end subroutine

end module
