module laplacian_mod

  use mem_mod

  implicit none

  logical, private                    :: zero_bound
  integer, private                            :: n1, n2, nt
  integer, private                            :: fat
  integer, private                            :: iz, ix, it
  real, private                               :: stencil0
  real, dimension(5), private                 :: stencilz, stencilx
  real, dimension(:,:), allocatable, private  :: p0, p1

  contains

  subroutine laplacian_init(n1_in, n2_in, nt_in, d1_in, d2_in,zero_bound_in)
    optional           :: zero_bound_in
    logical             :: zero_bound_in
    integer                        :: n1_in, n2_in, nt_in
    real                           :: d1_in, d2_in

    if(present(zero_bound_in)) then
      zero_bound=zero_bound_in
    else
      zero_bound=.false.
    end if

    n1 = n1_in
    n2 = n2_in
    nt = nt_in

!write(0,*)'In lapl init n1=',n1
!write(0,*)'In lapl init n2=',n2
!write(0,*)'In lapl init nt=',nt

    fat = 5

    stencilz(1)=3.333333333/(2.*d1_in*d1_in)
    stencilz(2)=-0.4761904762/(2.*d1_in*d1_in)
    stencilz(3)=0.0793650794/(2.*d1_in*d1_in)
    stencilz(4)=-0.0099206349/(2.*d1_in*d1_in)
    stencilz(5)=0.0006349206/(2.*d1_in*d1_in)

    stencilx(1)=3.333333333/(2.*d2_in*d2_in)
    stencilx(2)=-0.4761904762/(2.*d2_in*d2_in)
    stencilx(3)=0.0793650794/(2.*d2_in*d2_in)
    stencilx(4)=-0.0099206349/(2.*d2_in*d2_in)
    stencilx(5)=0.0006349206/(2.*d2_in*d2_in)

    stencil0=(stencilz(5)+stencilx(5))
    stencil0=stencil0+(stencilz(4)+stencilx(4))
    stencil0=stencil0+(stencilz(3)+stencilx(3))
    stencil0=stencil0+(stencilz(2)+stencilx(2))
    stencil0=stencil0+(stencilz(1)+stencilx(1))
    stencil0=-2.*stencil0

!write(0,*)'stencil0=',stencil0

    if(.not. allocated(p0)) then
      allocate(p0(1-fat:n1+fat,1-fat:n2+fat))
      call mem_add_4(size(p0))
    end if
    if(.not. allocated(p1)) then
      allocate(p1(n1,n2))
      call mem_add_4(size(p1))
    end if

!    if(.not. allocated(p1)) then
!      allocate(p1(1-fat:n1+fat,1-fat:n2+fat))
!      call mem_add_4(size(p1))
!    end if

  end subroutine

  function laplacian_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call laplacian_op2(adj, add, model, data)
    stat=0
  end function

  subroutine laplacian_op2(adj, add, model, data)
    logical,intent(in)        :: adj, add 
    real, dimension(n1,n2,nt) :: model, data

    p0 = 0.
    p1 = 0.

    if(adj) then
      if(.not. add) model = 0.
!write(0,*)"in lapl maxval(data)=",maxval(data)
      do it=1,nt
        p0(1:n1,1:n2) = data(:,:,it)
        if(zero_bound) then
!          model(fat+1:n1-fat,fat+1:n2-fat,it) =  &
!          model(fat+1:n1-fat,fat+1:n2-fat,it) +  &
!             p1(fat+1:n1-fat,fat+1:n2-fat)
          p0(1:fat,:) = 0
          p0(:,1:fat) = 0
          p0(n1-fat+1:n1,:) = 0
          p0(:,n2-fat+1:n2) = 0
!        else
!!          model(:,:,it) = model(:,:,it) + p1(1:n1,1:n2)
!!          model(:,:,it) = model(:,:,it) + p1(:,:)
!          p0(1:n1,1:n2) = data(:,:,it)
        end if
        p1 = 0
        do ix=1,n2
          do iz=1,n1
            p1(iz,ix) =   stencil0*p0(iz,ix)+&
                          stencilz(1)*(p0(iz-1,ix  )+p0(iz+1,ix  ))+&
                          stencilx(1)*(p0(iz  ,ix-1)+p0(iz  ,ix+1))+&
                          stencilz(2)*(p0(iz-2,ix  )+p0(iz+2,ix  ))+&
                          stencilx(2)*(p0(iz  ,ix-2)+p0(iz  ,ix+2))+&
                          stencilz(3)*(p0(iz-3,ix  )+p0(iz+3,ix  ))+&
                          stencilx(3)*(p0(iz  ,ix-3)+p0(iz  ,ix+3))+&
                          stencilz(4)*(p0(iz-4,ix  )+p0(iz+4,ix  ))+&
                          stencilx(4)*(p0(iz  ,ix-4)+p0(iz  ,ix+4))+&
                          stencilz(5)*(p0(iz-5,ix  )+p0(iz+5,ix  ))+&
                          stencilx(5)*(p0(iz  ,ix-5)+p0(iz  ,ix+5))
          end do
        end do
          model(:,:,it) = model(:,:,it) + p1(:,:)
      end do
!write(0,*)"in lapl maxval(mode)=",maxval(model)
    else
      if(.not. add) data = 0.
      do it=1,nt
        p0(1:n1,1:n2) = model(:,:,it)
        p1 = 0
        do ix=1,n2
          do iz=1,n1
            p1(iz,ix) =   stencil0*p0(iz,ix)+&
                          stencilz(1)*(p0(iz-1,ix  )+p0(iz+1,ix  ))+&
                          stencilx(1)*(p0(iz  ,ix-1)+p0(iz  ,ix+1))+&
                          stencilz(2)*(p0(iz-2,ix  )+p0(iz+2,ix  ))+&
                          stencilx(2)*(p0(iz  ,ix-2)+p0(iz  ,ix+2))+&
                          stencilz(3)*(p0(iz-3,ix  )+p0(iz+3,ix  ))+&
                          stencilx(3)*(p0(iz  ,ix-3)+p0(iz  ,ix+3))+&
                          stencilz(4)*(p0(iz-4,ix  )+p0(iz+4,ix  ))+&
                          stencilx(4)*(p0(iz  ,ix-4)+p0(iz  ,ix+4))+&
                          stencilz(5)*(p0(iz-5,ix  )+p0(iz+5,ix  ))+&
                          stencilx(5)*(p0(iz  ,ix-5)+p0(iz  ,ix+5))
          end do
        end do
        if(zero_bound) then
!          data(1:fat,:,it) =  data(1:fat,:,it)
!          data(:,1:fat,it) =  data(:,1:fat,it)
!          data(n1-fat+1:n1,:,it) = data(n1-fat+1:n1,:,it) =  
!          data(:,n2-fat+1:n2,it) =  data(:,1:fat,it)

          data(fat+1:n1-fat,fat+1:n2-fat,it) =  &
          data(fat+1:n1-fat,fat+1:n2-fat,it) +  &
            p1(fat+1:n1-fat,fat+1:n2-fat)
        else
!          data(:,:,it) = data(:,:,it) + p1(1:n1,1:n2)
          data(:,:,it) = data(:,:,it) + p1(:,:)
        end if
      end do
    end if
  end subroutine

end module
