module secder_mod

  implicit none

  logical, private                    :: zero_bound
  integer, private                    :: n1, n2, nt, it
  real, private                       :: idt2

  contains

  subroutine secder_init(n1_in, n2_in, nt_in, dt_in,zero_bound_in)
    optional           :: zero_bound_in
    logical             :: zero_bound_in
    integer                      :: n1_in, n2_in, nt_in
    real                         :: dt_in

    if(present(zero_bound_in)) then
      zero_bound=zero_bound_in
    else
      zero_bound=.false.
    end if
     
    n1 = n1_in
    n2 = n2_in
    nt = nt_in
    idt2 = 1./dt_in/dt_in
!    write(0,*)'idt2 =',idt2
  end subroutine

  function secder_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call secder_op2(adj, add, model, data)
    stat=0
  end function

  subroutine secder_op2(adj, add, model, data)
    logical,intent(in)          :: adj, add 
    real, dimension(n1,n2,nt) :: model, data
!    write(0,*)'zero_bound=',zero_bound
    if(adj) then
      if(.not. add) model = 0.
      if(zero_bound) then
        data(:,:,1) = 0
        data(:,:,nt) = 0
      end if
      model(:,:,1) = model(:,:,1) + (data(:,:,2)-data(:,:,1))*idt2
      model(:,:,nt) = model(:,:,nt) + (data(:,:,nt-1)-data(:,:,nt))*idt2
      model(:,:,2:nt-1) = model(:,:,2:nt-1) + (data(:,:,1:nt-2)-2.*data(:,:,2:nt-1)+data(:,:,3:nt))*idt2
    else
      if(.not. add) data = 0.
      if(zero_bound) then
!        data(:,:,1) = data(:,:,1) 
!        data(:,:,nt) = data(:,:,nt) 
      else
        data(:,:,1) = data(:,:,1) + (model(:,:,2)-model(:,:,1))*idt2
        data(:,:,nt) = data(:,:,nt) + (model(:,:,nt-1)-model(:,:,nt))*idt2
      end if
      data(:,:,2:nt-1) = data(:,:,2:nt-1) + (model(:,:,1:nt-2)-2.*model(:,:,2:nt-1)+model(:,:,3:nt))*idt2
    end if
  end subroutine

end module
