module wave_oper_mod

  use mem_mod
  use secder_mod
  use laplacian_mod

  implicit none

  logical, private                    :: zero_bound
  integer, private                            :: n1, n2, nt
  integer, private                            :: it
  real, dimension(:,:), pointer, private      :: vel
  real, dimension(:,:,:), allocatable, private  :: temp1,temp2

  contains

  subroutine wave_oper_init(n1_in, n2_in, nt_in, vel_in,zero_bound_in)
    optional           :: zero_bound_in
    logical             :: zero_bound_in
    integer                      :: n1_in, n2_in, nt_in
    real, dimension(:,:), target   :: vel_in
    n1 = n1_in
    n2 = n2_in
    nt = nt_in

    if(present(zero_bound_in)) then
      zero_bound=zero_bound_in
    else
      zero_bound=.false.
    end if

!write(0,*)"In wave_oper_init: minval(vel_in)=",minval(vel_in)
!write(0,*)"In wave_oper_init: maxval(vel_in)=",maxval(vel_in)

    vel => vel_in

!write(0,*)"In wave_oper_init: minval(vel)=",minval(vel)
!write(0,*)"In wave_oper_init: maxval(vel)=",maxval(vel)

    if(.not. allocated(temp1)) then
      call mem_alloc(temp1, n1, n2, nt)
      temp1=0
    end if
    if(.not. allocated(temp2)) then
      call mem_alloc(temp2, n1, n2, nt)
      temp2=0
    end if

!    write(0,*)'idt2 =',idt2
  end subroutine

  function wave_oper_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call wave_oper_op2(adj, add, model, data)
    stat=0
  end function

  subroutine wave_oper_op2(adj, add, model, data)
    logical,intent(in)        :: adj, add 
    real, dimension(n1,n2,nt) :: data, model

    if(adj) then
      if(.not. add) model = 0.
!write(0,*)"minval(vel)=",minval(vel)
!write(0,*)"maxval(vel)=",maxval(vel)
      do it=1,nt
        temp2(:,:,it)=(vel(:,:)**2)*data(:,:,it)
      end do
      call laplacian_op2(.true., .false., temp1, temp2) 
      temp2=0
      call secder_op2(.true., .false., temp2, data) 
!      do it=1,nt
!        temp1(:,:,it)=(vel(:,:)**2)*temp1(:,:,it)
!      end do
!write(0,*)"In wave_oper: minval(temp1)=",minval(temp1)
!write(0,*)"In wave_oper: maxval(temp1)=",maxval(temp1)
!write(0,*)"In wave_oper: minval(model)=",minval(model)
!write(0,*)"In wave_oper: maxval(model)=",maxval(model)
      if(zero_bound) then
        model(6:n1-5,6:n2-5,2:nt-1)=model(6:n1-5,6:n2-5,2:nt-1)+ &
           (temp1(6:n1-5,6:n2-5,2:nt-1)-temp2(6:n1-5,6:n2-5,2:nt-1))
      else
        model=model+(temp1-temp2)
      end if
    else 
      if(.not. add) data = 0.
!write(0,*)"In wave_oper: minval(vel)=",minval(vel)
!write(0,*)"In wave_oper: maxval(vel)=",maxval(vel)
      call laplacian_op2(.false., .false.,model, temp1) 
!write(0,*)"In wave_oper: minval(temp1)=",minval(temp1)
!write(0,*)"In wave_oper: maxval(temp1)=",maxval(temp1)
      do it=1,nt
        temp1(:,:,it)=(vel(:,:)**2)*temp1(:,:,it)
      end do
      call secder_op2(.false., .false., model, temp2) 
!write(0,*)"In wave_oper: minval(temp2)=",minval(temp2)
!write(0,*)"In wave_oper: maxval(temp2)=",maxval(temp2)
      if(zero_bound) then
        data(6:n1-5,6:n2-5,2:nt-1)=data(6:n1-5,6:n2-5,2:nt-1)+ &
            (temp1(6:n1-5,6:n2-5,2:nt-1)-temp2(6:n1-5,6:n2-5,2:nt-1))
      else
        data=data+(temp1-temp2)
      end if
!do it=1,nt
!  call snap("Lapl.H",n1,n2,temp1(:,:,it))
!  call snap("SecDer.H",n1,n2,temp2(:,:,it))
!end do
    end if
  end subroutine

end module
