##############################################################
#Makefile to run multiscale FWI on random model perturbation
##############################################################
#Common PBS parameters
PBS_TMPL=${PROPAG}scripts/pbs_tmpl_mazama.sh
queue=sep
queue_cap=50
total_jobs=100
#Background model (Already padded)
${D}/background.H:
	cd AllScaleModel;\
	make Vel0-Grad.H; cd .. ; \
	cp AllScaleModel/Vel0-Grad.H $@

#Big model perturbation (Already padded)
${D}/big_pert.H:
	cd AllScaleModelNew;\
	make DVel-Rand-Grad.H
	cp AllScaleModelNew/DVel-Rand-Grad.H $@

#Small model perturbation (Already padded)
${D}/tiny_pert.H:
	cd AllScaleModelNew;\
	make DVel-Rand-tiny-Grad.H
	cp AllScaleModelNew/DVel-Rand-tiny-Grad.H $@

${D}/%_model.H: ${D}/background.H ${D}/%.H
	Add $^ >$@

#Wavelet to be used for modeling
${D}/wavelet.H:
	cd AllScaleModel;\
	make Wavelet-Grad.H
	Window3d <AllScaleModel/Wavelet-Grad.H max1=12 > $@
	
	
${D}/wavelet_LF.H: ${B}/Wavelet.x
	${B}/Wavelet.x par=${P}/LF_wavelet.p | Scale rscale=1000000 > tmp.H
	echo "d1=.006 tdelay=2.7" >>tmp.H
	Window3d < tmp.H max1=6 >$@
	/opt/SEP/SEP7.0/bin/Rm tmp.H

########################################################################
#Genereting observed data
########################################################################
#Orginal source positions' file: AllScaleModel/Par/Sou-Array-Grad.p
nshots=$(shell awk 'NR==2' Par/source_positions.p | sed -e 's/^[ \t]*//')
${P}/source_positions.p: ${D}/background.H ${B}/Geo_uniform.x
	Window3d <$< f2=149 max2=3570 > tmp_mod.H
	${B}/Geo_uniform.x <tmp_mod.H depth=0. dx=150. file=$@ >/dev/null
	/opt/SEP/SEP7.0/bin/Rm tmp_mod.H

#Orginal receiver positions' file: AllScaleModel/Par/Rec-Array-Grad.p
${P}/rec_positions.p: ${D}/background.H ${B}/Geo_uniform.x
	Window3d <$< max2=5080 > tmp_mod.H
	${B}/Geo_uniform.x <tmp_mod.H depth=0. file=$@ >/dev/null
	/opt/SEP/SEP7.0/bin/Rm tmp_mod.H

pressure_%.H: solver_prog ${B}/Propagator_hess_fw.x ${D}/wavelet.H ${P}/rec_positions.p ${P}/source_positions.p ${D}/%_model.H
	mkdir -p wrk
	cp ${D}/wavelet.H wrk
	cp ${D}/$*_model.H wrk/vp_tmp$*.H
	${B}/bte_acoustic_fw.py pbs_template=${PBS_TMPL} queues=${queue} queues_cap=${queue_cap} path_out=wrk prefix=$* nshots=${nshots} nshots_perjob=2 vel=vp_tmp$*.H par_rec=../${P}/rec_positions.p par_sou=../${P}/source_positions.p wavelet=wavelet.H total_jobs_cap=${total_jobs} threads=24 dt=2.0000001E-03 mod_padded=1
	Cat3d `ls wrk/$*-*.H  | sort -V` axis=3 > $@
	rm wrk/vp_tmp$*.H


########################################################################
#Inversion section
########################################################################
SPWD=$(shell echo ${PWD} | sed 's/\//\\\//g')
${D}/inv_mask.H: ${D}/background.H
	Window3d <$< min1=0 | Add scale=0.0 > junk_inv_mask.H
	Math file1=junk_inv_mask.H exp="file1+1.0" | Pad beg1=200 >$@
	/opt/SEP/SEP7.0/bin/Rm junk_inv_mask.H

${D}/inv_mask_null.H: ${D}/background.H
	Math file1=${D}/background.H exp="file1*0.0+1.0" >$@

data_allscale_tiny_%hz.H: pressure_tiny_pert.H
	Bandpass <$^ flo=2 fhi=$* phase=1 | Bandpass flo=2 fhi=$* phase=1 >$@

data_allscale_big_%hz.H: pressure_big_pert.H
	Bandpass <$^ flo=2 fhi=$* phase=1 | Bandpass flo=2 fhi=$* phase=1 >$@

#Tiny perturbation
#BFGS all frequencies no mask
allscale_tiny_pert_bfgs: solver_prog ${D}/background.H pressure_tiny_pert.H ${D}/wavelet.H ${P}/rec_positions.p ${P}/source_positions.p ${B}/Propagator_hess_fw.x ${B}/Bornfast_hess_fw.x ${B}/Bornfast_hess_adj.x ${D}/inv_mask_null.H
	mkdir -p wrk
	cp ${D}/wavelet.H wrk/wavelet.H
	sed 's/#PBS -N jobname/#PBS -N bfgsall/' <${PBS_TMPL}  | sed 's/#PBS -o/#PBS -o ${SPWD}\/$@.log/' | sed 's/#PBS -e/#PBS -e ${SPWD}\/$@.log/' | sed 's/#PBS -d/#PBS -d ${SPWD}/' | sed 's/set CURRDIR=/set CURRDIR=${SPWD}/' > $@.sh
	sed -i 's/jg_n24_256_none_FEX_rd_a/jg_n16_128_none_FDR1_rd_a/' $@.sh
	#Changing operator files
	cp ${P}/nl_acoustic_fw.p   ${P}/nl_acoustic_fwall.p
	cp ${P}/lin_acoustic_fw.p  ${P}/lin_acoustic_fwall.p
	cp ${P}/lin_acoustic_adj.p ${P}/lin_acoustic_adjall.p
	sed -i 's/inv_mask.H/inv_mask_null.H/' ${P}/lin_acoustic_fwall.p
	sed -i 's/inv_mask.H/inv_mask_null.H/' ${P}/lin_acoustic_adjall.p
	echo "${B}/nl_solver.py fwd_nl_cmd_file=${P}/nl_acoustic_fwall.p fwd_cmd_file=${P}/lin_acoustic_fwall.p adj_cmd_file=${P}/lin_acoustic_adjall.p data=pressure_tiny_pert.H init_model=${D}/background.H inv_model=$@.H suffix=_fwi_tiny_bfgs iteration_movies=obj,model,gradient wrk_dir=BFGS_ACOUSTIC_TINY niter=200 solver=lbfgs" >> $@.sh
	qsub $@.sh

BFGS_ACOUSTIC_TINY/model_fwi_tiny_bfgs.H: allscale_tiny_pert_bfgs
